package command

import (
	"github.com/open-io/oio-exporter/common"
	"log"
	"os/exec"
	"strings"
	"time"
)

type daemon struct {
	services **map[string][]*common.Service
	interval time.Duration
	commands []string
	versions map[string]string
}

var versionCommandsRPM = map[string]string{
	"openio":        "rpm -q --qf '%{VERSION}\n' openio-sds-server",
	"swift":         "rpm -q --qf '%{VERSION}\n' openio-sds-swift",
	"s3":            "rpm -q --qf '%{VERSION}\n' openio-sds-swift-plugin-swift3",
	"zookeeper":     "rpm -q --qf '%{VERSION}\n' zookeeper",
	"zookeeper_lib": "rpm -q --qf '%{VERSION}\n' libzookeeper2",
	"oiofs":         "rpm -q --qf '%{VERSION}\n' oiofs-fuse",
	"redis":         "redis-server --version | grep -oP ' v=\\K.+? '",
	"beanstalkd":    "beanstalkd -v | awk 'BEGIN {err=1} NF == 2 {print $2; err = 0} END {exit err}'",
	"keystone":      "keystone-manage --version 2>&1",
}

var versionCommandsDeb = map[string]string{
	"openio":        "dpkg-query -Wf='${Version}\n' openio-sds",
	"swift":         "dpkg-query -Wf='${Version}\n' openio-sds-swift",
	"swift3":        "dpkg-query -Wf='${Version}\n' openio-sds-swift-plugin-s3",
	"zookeeper":     "dpkg-query -Wf='${Version}\n' zookeeper",
	"zookeeper_lib": "dpkg-query -Wf='${Version}\n' libzookeeper-java",
	"oiofs":         "dpkg-query -Wf='${Version}\n' oiofs-fuse",
	"redis":         "redis-server --version | grep -oP ' v=\\K.+? '",
	"beanstalkd":    "beanstalkd -v | awk 'BEGIN {err=1} NF == 2 {print $2; err = 0} END {exit err}'",
	"keystone":      "keystone-manage --version 2>&1",
}

func MakeDaemon(services **map[string][]*common.Service, commands []string, interval time.Duration) *daemon {
	d := daemon{
		services: services,
		interval: interval,
		commands: []string{},
		versions: map[string]string{},
	}
	if checkCMD("rpm") {
		d.versions = versionCommandsRPM
	} else if checkCMD("dpkg-query") {
		d.versions = versionCommandsDeb
	} else {
		log.Println("WARN", "in determining distro", "version info will not be available")
	}
	return &d
}

func (d *daemon) Run(m *common.MetricData) {
	for serviceType, cmd := range d.versions {
		out, err := d.runCommand(cmd)
		if err != nil {
			m.IncrError("command", serviceType)
			log.Println("WARN", "in running command", cmd, serviceType, err)
			continue
		}
		m.Set(&common.Metric{
			Name: "version",
			Labels: [][]string{
				[]string{"service_type", serviceType},
				[]string{"version", out},
			},
			Value: float64(time.Now().Unix()),
		})
	}
	time.Sleep(d.interval)
}

func checkCMD(cmd string) bool {
	_, err := exec.LookPath(cmd)
	return err == nil
}

func (d *daemon) runCommand(cmd string) (string, error) {
	out, err := exec.Command("/bin/bash", "-c", cmd).Output()
	if err != nil {
		return "", err
	}
	outFmt := strings.Trim(
		strings.TrimSuffix(string(out), "\n"), " ")

	return outFmt, nil
}
