package main

import (
	"flag"
	"fmt"
	cmd "github.com/open-io/oio-exporter/command"
	"github.com/open-io/oio-exporter/common"
	hc "github.com/open-io/oio-exporter/healthcheck"
	self "github.com/open-io/oio-exporter/internal"
	clog "github.com/open-io/oio-exporter/log"
	"github.com/open-io/oio-exporter/oiofs"
	"github.com/open-io/oio-exporter/openio"
	"github.com/open-io/oio-exporter/redis"
	"github.com/open-io/oio-exporter/s3rt"
	"github.com/open-io/oio-exporter/tcp"
	"gopkg.in/yaml.v2"
	"io"
	"log"
	"net/http"
	"os"
	"strings"
	"time"
)

var HEALTHCHECKS = ""
var LOGPATTERNS = ""

type inventory struct {
	Namespaces map[string]struct {
		Services map[string][]*common.Service `yaml:"services"`
	} `yaml:"namespaces"`
}

type oioExporter struct {
	bind       string
	inventory  string
	services   *map[string][]*common.Service
	targets    *[]*common.Service
	ns         string
	hcTargets  []string
	metrics    *common.MetricData
	lastUpdate *time.Time
}

// Export prometheus metrics
func (exp *oioExporter) export(w http.ResponseWriter, r *http.Request) {
	filter := r.URL.Query().Get("filter")
	for _, metric := range exp.metrics.Measures {
		if filter != "" && !strings.HasPrefix(metric.Name, filter) {
			continue
		}
		labelStr := []string{}
		for _, labelPair := range metric.Labels {
			labelStr = append(labelStr, labelPair[0]+"=\""+labelPair[1]+"\"")
		}

		if metric.Accumulator != nil {
			for k, v := range metric.Accumulator.Values() {
				fmt.Fprintf(w, "%s{%s} %f\n", metric.Name+"_"+k, strings.Join(labelStr, ","), v)
			}
		} else {
			fmt.Fprintf(w, "%s{%s} %f\n", metric.Name, strings.Join(labelStr, ","), metric.Value)
		}
	}
}

// Reload local service configuration
func (exp *oioExporter) reload(w http.ResponseWriter, r *http.Request) {
	n, err := exp.loadConfig(true)
	if err != nil {
		log.Fatalln(err)
	}
	log.Println("Reloaded", n, "local targets")
	fmt.Fprintf(w, "OK")
}

// Serve local service configuration
func (exp *oioExporter) config(w http.ResponseWriter, r *http.Request) {
	file, err := os.Open(exp.inventory)
	if err != nil {
		log.Fatal(err)
	}
	if _, err := io.Copy(w, file); err != nil {
		log.Println("WARN", "could not serve config", err)
	}
}

func (exp *oioExporter) loadConfig(local bool) (n int, err error) {
	if local {
		file, err := os.Open(exp.inventory)
		if err != nil {
			return n, err
		}
		defer file.Close()
		ns, services, err := exp.parseConfig(file)
		for svcType, inst := range services {
			for _, svc := range inst {
				svc.Type = svcType
			}
		}
		exp.services = &services
		exp.ns = ns
		n = len(services)
		lastUpdate := time.Now()
		exp.lastUpdate = &lastUpdate
		return n, err
	}

	services := []*common.Service{}

	for _, target := range exp.hcTargets {
		var res *http.Response
		res, err = http.Get("http://" + target + "/config")
		if err != nil {
			return n, err
		}
		var remoteServices map[string][]*common.Service
		_, remoteServices, err = exp.parseConfig(res.Body)
		if err != nil {
			return
		}
		for svcType, inst := range remoteServices {
			for _, svc := range inst {
				svc.Type = svcType
				services = append(services, svc)
			}
			n++
		}
	}

	// TODO: lock
	exp.targets = &services
	log.Println("setting targets", len(services))
	return
}

func (exp *oioExporter) parseConfig(r io.Reader) (ns string, svc map[string][]*common.Service, err error) {
	var inv inventory
	if err = yaml.NewDecoder(r).Decode(&inv); err != nil {
		return
	}
	for ns, svc := range inv.Namespaces {
		return ns, svc.Services, nil
	}
	return
}

func (exp *oioExporter) run() {
	http.HandleFunc("/metrics", exp.export)
	http.HandleFunc("/config", exp.config)
	http.HandleFunc("/reload", exp.reload)
	log.Fatalln(http.ListenAndServe(exp.bind, nil))
}

func (exp *oioExporter) reloadRemote() {
	for {
		// Load remote configuration
		n2, err := exp.loadConfig(false)
		retry := 10 * time.Second
		if err != nil {
			log.Println("WARN", "in loading remote config", err, "retrying in", retry)
			time.Sleep(retry)
			continue
		}
		log.Println("(Re)loaded", n2, "remote targets")
		break
	}
}

func main() {
	var inventory string
	var targets string
	var patternPath string
	var bind string
	fs := flag.NewFlagSet("", flag.ExitOnError)
	fs.StringVar(&inventory, "inventory", "/etc/oio/sds/OPENIO/inventory.yml", "Path to inventory file")
	fs.StringVar(&patternPath, "pattern-path", "/etc/oio-exporter-patterns.yml", "Path to log patterns")
	fs.StringVar(&bind, "bind", "0.0.0.0:9115", "Listen address")
	fs.StringVar(&targets, "targets", "localhost:9115", "Comma separated list of targets to perform healthchecks on")
	if err := fs.Parse(os.Args[1:]); err != nil {
		log.Fatalln("Could not parse args", err)
	}

	services := []*common.Service{}
	lastUpdate := time.Unix(0, 0)
	exp := oioExporter{
		bind:       bind,
		inventory:  inventory,
		hcTargets:  strings.Split(targets, ","),
		metrics:    common.MakeMetricData(),
		targets:    &services,
		lastUpdate: &lastUpdate,
	}

	// Load local configuration
	n1, err := exp.loadConfig(true)
	if err != nil {
		log.Fatalln("FATAL", "in loading local config", err)
	}
	log.Println("Loaded", n1, "local targets")

	go exp.reloadRemote()
	hcDaemon := hc.MakeDaemon(HEALTHCHECKS, &exp.targets, 5*time.Second)
	oioDaemon := openio.MakeDaemon(exp.ns, &exp.services, 5*time.Second)
	redisDaemon := redis.MakeDaemon(&exp.services, 5*time.Second)
	tcpDaemon := tcp.MakeDaemon(&exp.services, 5*time.Second)
	cmdDaemon := cmd.MakeDaemon(&exp.services, []string{}, 5*time.Second)
	s3rtDaemon := s3rt.MakeDaemon(&exp.services, 5*time.Second)
	logDaemon := clog.MakeDaemon(LOGPATTERNS, patternPath, &exp.services, 5*time.Second)
	oiofsDaemon := oiofs.MakeDaemon(&exp.services, 5*time.Second)
	selfDaemon := self.MakeDaemon(5 * time.Second)

	go hcDaemon.Run(exp.metrics, false)
	go oioDaemon.Run(exp.metrics)
	go redisDaemon.Run(exp.metrics, &exp.lastUpdate)
	go tcpDaemon.Run(exp.metrics)
	go cmdDaemon.Run(exp.metrics)
	go s3rtDaemon.Run(exp.metrics)
	go logDaemon.Run(exp.metrics)
	go oiofsDaemon.Run(exp.metrics)
	go selfDaemon.Run(exp.metrics)

	exp.run()
}
