package s3rt

import (
	"bytes"
	"context"
	"crypto/md5"
	"fmt"
	"github.com/open-io/oio-exporter/common"
	"log"
	"net/http"
	"strconv"
	"time"

	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/awserr"
	"github.com/aws/aws-sdk-go/aws/credentials"
	"github.com/aws/aws-sdk-go/aws/session"
	"github.com/aws/aws-sdk-go/service/s3"
	"github.com/aws/aws-sdk-go/service/s3/s3iface"
	"github.com/aws/aws-sdk-go/service/s3/s3manager"
)

type s3c struct {
	s3       s3iface.S3API
	bucketRT bool
	service  string
	bucket   string
}

type daemon struct {
	services **map[string][]*common.Service
	interval time.Duration
	clients  []*s3c
	data     *[]byte
	errors   []*common.Service
}

const multMs = 10e6
const mpuSize = 5 * 1024 * 1024
const objSize = mpuSize + 1

func MakeDaemon(services **map[string][]*common.Service, interval time.Duration) *daemon {
	data := make([]byte, objSize)

	d := daemon{
		services: services,
		interval: interval,
		clients:  []*s3c{},
		data:     &data,
		errors:   []*common.Service{},
	}

	if svces, ok := (**services)["oioswift"]; ok {
		for _, svc := range svces {
			valid := true
			for _, k := range []string{"access", "secret", "region"} {
				if _, ok := svc.Conf[k]; !ok {
					log.Println("WARN", "S3 roundtrip", "missing config key in inventory, will not collect", k, svc)
					d.errors = append(d.errors, svc)
					valid = false
					break
				}
			}
			if !valid {
				continue
			}
			bucketRT := false
			if v, ok := svc.Conf["bucket_roundtrip"]; ok && v == "true" {
				bucketRT = true
			}

			endpoint := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
			config := aws.Config{
				Region:           aws.String(svc.Conf["region"]),
				Credentials:      credentials.NewStaticCredentials(svc.Conf["access"], svc.Conf["secret"], ""),
				Endpoint:         aws.String("http://" + endpoint),
				S3ForcePathStyle: aws.Bool(true),
				MaxRetries:       aws.Int(1),
				HTTPClient:       &http.Client{Timeout: time.Second * interval},
			}
			sess, err := session.NewSession(&config)
			if err != nil {
				log.Println("WARN", "in initializing client", svc, err)
				continue
			}
			d.clients = append(d.clients, &s3c{
				s3:       s3.New(sess),
				bucketRT: bucketRT,
				service:  endpoint,
				bucket:   "s3rt-" + fmt.Sprintf("%x", md5.Sum([]byte(endpoint))),
			})
		}
	}
	return &d
}

func (d *daemon) Run(m *common.MetricData) {
	for _, svc := range d.errors {
		m.IncrError("s3rt", svc.Type)
	}
	for _, client := range d.clients {
		go func(c *s3c) {
			var t time.Duration
			var err error
			doBucketCreate := true
			for {
				if doBucketCreate || c.bucketRT {
					t, err = c.mb(c.bucket)
					if err == nil {
						doBucketCreate = false
					}
					if !c.bucketRT && err != nil {
						m.IncrError("s3rt", "s3rt")
						log.Println("WARN", "in precreating bucket", c, err)
					}
					if c.bucketRT {
						d.s3Publish(m, c.service, "mb", t, err, false)
					}
				}
				t, err = c.put(c.bucket, "s3roundtrip_object", d.data, mpuSize)
				d.s3Publish(m, c.service, "put", t, err, false)

				t, err = c.get(c.bucket, "s3roundtrip_object")
				d.s3Publish(m, c.service, "get", t, err, false)

				t, err = c.ls(c.bucket, 1000)
				d.s3Publish(m, c.service, "ls", t, err, false)

				t, err = c.del(c.bucket, "s3roundtrip_object")
				d.s3Publish(m, c.service, "del", t, err, false)

				// Time to first byte
				dataTtfb := []byte("0")
				t, err = c.put(c.bucket, "s3roundtrip_ttfb_object", &dataTtfb, mpuSize)
				d.s3Publish(m, c.service, "put", t, err, true)
				if err == nil {
					t, err = c.get(c.bucket, "s3roundtrip_ttfb_object")
					d.s3Publish(m, c.service, "get", t, err, true)

					_, _ = c.del(c.bucket, "s3roundtrip_ttfb_object")
				}

				if c.bucketRT {
					t, err = c.rb(c.bucket)
					d.s3Publish(m, c.service, "rb", t, err, false)
				}
				time.Sleep(d.interval)
			}
		}(client)
	}
}

func (d *daemon) s3Publish(m *common.MetricData, service, request string, time time.Duration, err error, ttfb bool) {
	code := "2xx"
	if err != nil {
		if req, ok := err.(awserr.RequestFailure); ok {
			// Error is an AWS request failure
			status := req.StatusCode() / 100
			if status == 3 || status == 1 {
				// Treat redirects and continues as 2xx
				status = 2
			}
			code = strconv.Itoa(status) + "xx"
		} else {
			code = "other"
		}
	}
	if ttfb {
		m.Set(&common.Metric{
			Name: "s3rt_ttfb",
			Labels: [][]string{
				[]string{"service", service},
				[]string{"request", request},
			},
			Value: float64(time.Nanoseconds() / multMs),
		})
	} else {
		m.Set(&common.Metric{
			Name: "s3rt_response_time",
			Labels: [][]string{
				[]string{"service", service},
				[]string{"request", request},
			},
			Value: float64(time.Nanoseconds() / multMs),
		})
		m.Set(&common.Metric{
			Name: "s3rt_response_code",
			Labels: [][]string{
				[]string{"service", service},
				[]string{"request", request},
				[]string{"code", code},
			},
			Value: float64(1),
		})
	}
}

func (c *s3c) put(bucket, object string, data *[]byte, mpuSize int64) (time.Duration, error) {
	uploader := s3manager.NewUploaderWithClient(c.s3)
	inputUpload := &s3manager.UploadInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(object),
		Body:   bytes.NewReader(*data),
	}
	ctx, cancel := context.WithTimeout(context.Background(), 30*time.Second)
	defer cancel()

	start := time.Now()
	_, err := uploader.UploadWithContext(ctx, inputUpload, func(u *s3manager.Uploader) {
		u.PartSize = mpuSize
	})
	return time.Since(start), err
}

func (c *s3c) get(bucket, object string) (time.Duration, error) {
	input := &s3.GetObjectInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(object),
	}
	start := time.Now()
	_, err := c.s3.GetObject(input)
	return time.Since(start), err
}

func (c *s3c) del(bucket, obj string) (time.Duration, error) {
	input := &s3.DeleteObjectInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(obj),
	}
	start := time.Now()
	_, err := c.s3.DeleteObject(input)
	return time.Since(start), err
}

func (c *s3c) mb(bucket string) (time.Duration, error) {
	input := s3.CreateBucketInput{
		Bucket: aws.String(bucket),
	}
	start := time.Now()
	_, err := c.s3.CreateBucket(&input)
	if aerr, ok := err.(awserr.Error); ok && aerr.Code() == s3.ErrCodeBucketAlreadyOwnedByYou {
		return time.Since(start), nil
	}
	return time.Since(start), err
}

func (c *s3c) rb(bucket string) (time.Duration, error) {
	input := s3.DeleteBucketInput{
		Bucket: aws.String(bucket),
	}
	start := time.Now()
	_, err := c.s3.DeleteBucket(&input)
	if err != nil {
		return time.Since(start), err
	}
	return time.Since(start), nil
}

func (c *s3c) ls(bucket string, keys int64) (time.Duration, error) {
	input := s3.ListObjectsInput{
		Bucket:  aws.String(bucket),
		MaxKeys: aws.Int64(keys),
	}
	start := time.Now()
	_, err := c.s3.ListObjects(&input)
	if err != nil {
		return time.Since(start), err
	}
	return time.Since(start), nil
}
